/*
 * smdb.c
 *
 * Copyright 2002, 2004 by Anthony Howe. All rights reserved.
 */

#ifndef SMCF_LINE_SIZE
#define SMCF_LINE_SIZE	1024
#endif

#ifndef SMDB_KEY_SIZE
#define SMDB_KEY_SIZE	512
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <com/snert/lib/version.h>

#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#ifdef __WIN32__
# if defined(__VISUALC__)
#  define _WINSOCK2API_
# endif
# include <windows.h>
# include <winsock2.h>
# include <com/snert/lib/io/Log.h>
#else
# include <syslog.h>
# include <sys/socket.h>
#endif

#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/mail/smdb.h>

static long debug = 0;

void
smdbSetDebugMask(int mask)
{
	debug = mask;
}

/***********************************************************************
 *** smdb open routines. These 3 functions are a critical section.
 ***********************************************************************/

#ifdef LIBSNERT_WITH_BERKELEY_DB

int
smdbReopen(smdb *sm)
{
	if (sm->db != (DB *) 0) {
		/* Even though we specified DB_READONLY for the DB->open(),
		 * DB->close() appears to generate a "permission denied"
		 * error, because the system wants to update the file
		 * access time (I'm guessing). So we ignore it.
		 */
#if DB_VERSION_MAJOR == 1
		(void) sm->db->close(sm->db);
#else
		(void) sm->db->close(sm->db, DB_NOSYNC);
#endif
	}
	
#if DB_VERSION_MAJOR == 1
	if ((sm->db = dbopen(sm->filename, O_RDONLY, 0, sm->dbtype, NULL)) == (DB *) 0) {
		syslog(LOG_ERR, "failed to open \"%s\": %s (%d)", sm->filename, strerror(errno), errno);
		goto error0;
	}
#else
{
	int rc;
	
	rc = db_create(&sm->db, (DB_ENV *) 0, 0);
	if (rc != 0 || sm->db == (DB *) 0) {
		syslog(LOG_ERR, "db_create() error: %s", db_strerror(rc));
		goto error0;
	}

	if (debug & SMDB_DEBUG_ALL) {
		sm->db->set_errpfx(sm->db, sm->filename);
		sm->db->set_errfile(sm->db, stderr);
	}

	rc = sm->db->verify(sm->db, sm->filename, (const char *) 0, (FILE *) 0, 0);
	if (rc != 0) {
		syslog(LOG_ERR, "\"%s\" failed verification: %s", sm->filename, db_strerror(rc));
		goto error1;
	}

	/* BDB 4.2.52 documentation:
	 *	
	 *	The DB handle may not be accessed again after
	 *	DB->verify is called, regardless of its return.
	 *
	 * In previous releases, applications calling the DB->verify
	 * method had to explicitly discard the DB handle by calling
	 * the DB->close method. Further, using the DB handle in other
	 * ways after calling the DB->verify method was not prohibited
	 * by the documentation, although such use was likely to lead
	 * to problems.
	 *
	 * For consistency with other Berkeley DB methods, DB->verify
	 * method has been documented in the current release as a DB
	 * handle destructor. Applications using the DB handle in any
	 * way (including calling the DB->close method) after calling
	 * DB->verify should be updated to make no further use of any
	 * kind of the DB handle after DB->verify returns.
	 */
#if DB_VERSION_MAJOR == 1
	(void) sm->db->close(sm->db);
#elif DB_VERSION_MAJOR < 4 || DB_VERSION_MINOR < 2
	(void) sm->db->close(sm->db, 0);
#endif

	rc = db_create(&sm->db, (DB_ENV *) 0, 0);
	if (rc != 0 || sm->db == (DB *) 0) {
		syslog(LOG_ERR, "db_create() error: %s", db_strerror(rc));
		goto error0;
	}

	if (debug & SMDB_DEBUG_ALL) {
		sm->db->set_errpfx(sm->db, sm->filename);
		sm->db->set_errfile(sm->db, stderr);
	}

	/* Use the same default cache size as Sendmail. */
	rc = sm->db->set_cachesize(sm->db, 0, 1024*1024, 0);
	if (rc != 0)
		syslog(LOG_WARNING, "warning: failed to set cache size for \"%s\": %s", sm->filename, db_strerror(rc));

	rc = sm->db->open(sm->db, DBTXN sm->filename, (const char *) 0, sm->dbtype, DB_RDONLY, 0);
	if (rc != 0) {
		syslog(LOG_ERR, "failed to open \"%s\": %s", sm->filename, db_strerror(rc));
		goto error1;
	}
}
#endif

	if (stat(sm->filename, &sm->finfo) != 0) {
		sm->finfo.st_mtime = 0;
		sm->finfo.st_mode = 0;
	}

	return 0;
#if DB_VERSION_MAJOR == 1
	(void) sm->db->close(sm->db);
#else
error1:
	(void) sm->db->close(sm->db, 0);
#endif
error0:	
	return -1;
}

smdb *
smdbOpen(const char *filename, int dbtype)
{
	smdb *sm;

	/* If name is null or empty, then disable use of the database. */
	if (filename == (const char *) 0 || filename[0] == '\0')
		goto error0;

	if ((sm = (smdb *) calloc(1, sizeof(*sm))) == (smdb *) 0)
		goto error0;
		
	if ((sm->filename = strdup(filename)) == (char *) 0)
		goto error1;

	sm->dbtype = dbtype;
	
	if (smdbReopen(sm))
		goto error2;

	return sm;
error2:
	free(sm->filename);
error1:
	free(sm);
error0:
	return (smdb *) 0;
}

int
smdbReload(smdb *sm)
{
	struct stat finfoNow;
	
	if (sm == (smdb *) 0)
		return -1;

	if (stat(sm->filename, &finfoNow) == 0 && sm->finfo.st_mtime != finfoNow.st_mtime) {
		if (debug & (SMDB_DEBUG_GET|SMDB_DEBUG_SET))
			syslog(LOG_DEBUG, "sendmail db \"%s\" updated, reopening", sm->filename);

		if (smdbReopen(sm)) {
			syslog(LOG_CRIT, "failed to re-open \"%s\"", sm->filename);
			return -1;
		}
	}
	
	return 0;
}

/***********************************************************************
 *** Common Routines
 ***********************************************************************/

int
smdbGetValue(smdb *sm, DBT *key, DBT *value)
{
	int rc;

	if (sm == (smdb *) 0 || sm->db == (DB *) 0 || key == (DBT *) 0 || value == (DBT *) 0)
		return -1;		

	if (debug & SMDB_DEBUG_GET)
		syslog(LOG_DEBUG, "checking \"%s\" for \"%s\"", sm->filename, (char *) key->data);

#if DB_VERSION_MAJOR == 1
	rc = sm->db->get(sm->db, key, value, 0);
#else
	rc = sm->db->get(sm->db, (DB_TXN *) 0, key, value, 0);
#endif

	if (rc == DB_NOTFOUND)
		return -1;

	if (rc != 0) {
		syslog(LOG_ERR, "sendmail db get \"%s\" error: %s", sm->filename, db_strerror(rc));
		return -1;
	}

	return 0;
}

/***********************************************************************
 *** smdbAliases Routines
 ***********************************************************************/

int
smdbAliasesHasUser(Vector smdbAliasesList, char *user)
{
	int i;
	DBT key, value;

	if (smdbAliasesList == (Vector) 0)
		return 0;

	/*** NOTE alaises database keys _include_ the null byte as part of
	 *** the key, which is different from the access database keys.
	 ***/
	memset(&key, 0, sizeof (key));
	key.size = strlen(user) + 1;
	key.data = user;

	for (i = 0; i  < VectorLength(smdbAliasesList); i++) {
		smdb *entry = VectorGet(smdbAliasesList, i);
		if (entry == (smdb *) 0) {
			syslog(LOG_CRIT, "null pointer for aliases database structure, index %d", i);
			return -1;
		}

		memset(&value, 0, sizeof (value));

#if DB_VERSION_MAJOR > 1
		value.flags = DB_DBT_PARTIAL|DB_DBT_USERMEM;
		value.ulen = 0;
		value.dlen = 0;
		value.doff = 0;
#endif

		if (smdbGetValue(entry, &key, &value) == 0) {
			return 1;
		}
	}

	return 0;
}

/***********************************************************************
 *** smdbAccess Routines
 ***********************************************************************/

/*
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessGetAccess(smdb *sm, DBT *key)
{
	DBT value;
	int status;
	char text[8];
	
	if (sm == (smdb *) 0 || sm->db == (DB *) 0 || key == (DBT *) 0)
		return SMDB_ACCESS_UNKNOWN;		

	memset(&text, 0, sizeof (text));
	memset(&value, 0, sizeof (value));

#if DB_VERSION_MAJOR > 1
	value.flags = DB_DBT_PARTIAL|DB_DBT_USERMEM;
	value.ulen = sizeof (text)-1;
	value.dlen = sizeof (text)-1;
	value.doff = 0;
	value.data = text;
#endif

	if (smdbGetValue(sm, key, &value))
		return SMDB_ACCESS_UNKNOWN;

	if (value.size <= 0)
		return SMDB_ACCESS_UNKNOWN;

#if DB_VERSION_MAJOR == 1
{
	long n = value.size < sizeof (text) ? value.size : sizeof (text)-1;
	(void) memcpy(text, value.data, n);
	text[n] = '\0';		
}
#endif
	if (debug & SMDB_DEBUG_GET)
		syslog(LOG_DEBUG, "access DB key=\"%s\" leading value=\"%s\"", (char *) key->data, text);

	if (text[0] == SMDB_ACCESS_OK || text[0] == SMDB_ACCESS_SKIP || text[0] == SMDB_ACCESS_FRIEND || text[2] == SMDB_ACCESS_RELAY) {
		status = SMDB_ACCESS_OK;
	} else if (text[0] == SMDB_ACCESS_DISCARD) {
		/* Technically a DISCARD is an accept and silently
		 * delete without an error response.
		 */
		status = SMDB_ACCESS_DISCARD;
	} else if (text[0] == SMDB_ACCESS_ERROR || text[0] == SMDB_ACCESS_HATER || text[2] == SMDB_ACCESS_REJECT || isdigit(text[0])) {
		/* Other possible results are REJECT, ERROR, ### */
		status = SMDB_ACCESS_REJECT;
	} else {
		syslog(LOG_WARNING, "access DB key=\"%s\" contains unknown leading value=\"%s\"", (char *) key->data, text);
		status = SMDB_ACCESS_UNKNOWN;
	}

	return status;
}

/*
 * Lookup
 *
 *	tag:a.b.c.d
 *	tag:a.b.c
 *	tag:a.b
 *	tag:a
 *
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessGetIpAccess(smdb *sm, const char *tag, const char *ip)
{
	DBT key;
	int status;
	char *kbuffer;

	status = SMDB_ACCESS_UNKNOWN;

	if (sm == (smdb *) 0 || ip == (const char *) 0)
		goto error0;		

	if ((kbuffer = malloc(SMDB_KEY_SIZE)) == (char *) 0)
		goto error0;		

	if (tag == (const char *) 0)
		tag = "";
		
	memset(&key, 0, sizeof (key));
	
	/* Note the trailing senteniel dot is removed on the first 
	 * pass through the loop.
	 */
	key.size = snprintf(kbuffer, SMDB_KEY_SIZE, "%s%s.", tag, ip);
	key.data = kbuffer;
	
	while (0 < key.size--) {
		if (kbuffer[key.size] == '.') {
			kbuffer[key.size] = '\0';
			if ((status = smdbAccessGetAccess(sm, &key)) != SMDB_ACCESS_UNKNOWN) {
				/* Got a result! */
				break;
			}
		}
	}

	free(kbuffer);
error0:
	return status;
}

/*
 * Lookup
 *
 *	tag:[ip]
 *	tag:[ipv6:ip]
 *	tag:some.sub.domain.tld
 *	tag:sub.domain.tld
 *	tag:domain.tld
 *	tag:tld
 *
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessGetDomainAccess(smdb *sm, const char *tag, const char *domain)
{
	DBT key;
	long tagLength;
	int status, resolved;
	char *copyHere, *kbuffer;

	status = SMDB_ACCESS_UNKNOWN;
	
	if (sm == (smdb *) 0 || domain == (const char *) 0)
		goto error0;		

	if ((kbuffer = malloc(SMDB_KEY_SIZE)) == (char *) 0)
		goto error0;		

	/* If the domain didn't resolve, then its an ip as domain name form
	 * so we only want to do one lookup on the whole and avoid the parent
	 * domain lookups.
	 */
	resolved = domain[0] != '[';

	memset(&key, 0, sizeof (key));

	if (tag == (char *) 0) {
		tagLength = 0;
		copyHere = kbuffer;
	} else {
		/* Note that tag must include colon separator. */
		tagLength = snprintf(kbuffer, SMDB_KEY_SIZE, "%s", tag);

		/* This should never happen. */
		if (SMDB_KEY_SIZE <= tagLength)
			goto error1;

		copyHere = kbuffer + tagLength;
	}

	/* Assume that the domain starts with a leading dot (.) */
	domain -= 1;

	do {
		(void) strncpy(copyHere, domain+1, SMDB_KEY_SIZE - tagLength);
		key.size = strlen(kbuffer);
		key.data = kbuffer;

		/* Remove trailing (root) dot just before the '\0' from domain name. */
		if (1 < key.size && kbuffer[key.size - 1] == '.')
			kbuffer[--key.size] = '\0';

		if ((status = smdbAccessGetAccess(sm, &key)) != SMDB_ACCESS_UNKNOWN) {
			/* Got a result! */
			break;
		}
	} while (resolved && (domain = strchr(domain+1, '.')) != (const char *) 0 && domain[1] != '\0');
error1:
	free(kbuffer);
error0:
	return status;
}

/*
 * Lookup
 *
 *	tag:account@some.sub.domain.tld
 *	tag:some.sub.domain.tld
 *	tag:sub.domain.tld
 *	tag:domain.tld
 *	tag:tld
 *	tag:account@
 *
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessGetEmailDomainUserAccess(smdb *sm, const char *tag, const char *email, const char *domain, const char *user)
{
	DBT key;
	int status;
	char *kbuffer;

	status = SMDB_ACCESS_UNKNOWN;

	if (sm == (smdb *) 0)
		goto error0;		

	if ((kbuffer = malloc(SMDB_KEY_SIZE)) == (char *) 0)
		goto error0;		

	memset(&key, 0, sizeof (key));

	if (tag == (const char *) 0)
		tag = "";

	/* Lookup Tag:sender */
	if (email != (const char *) 0) {
		key.size = snprintf(kbuffer, SMDB_KEY_SIZE, "%s%s", tag, email);
		TextLower(kbuffer, 0, -1);
		key.data = kbuffer;

		if ((status = smdbAccessGetAccess(sm, &key)) != SMDB_ACCESS_UNKNOWN)
			goto result;
	}
	
	/* Lookup Tag:domain */
	if ((status = smdbAccessGetDomainAccess(sm, tag, domain)) != SMDB_ACCESS_UNKNOWN)
		goto result;
	
	/* Lookup Tag:user@ */
	if (user != (const char *) 0) {
		key.size = snprintf(kbuffer, SMDB_KEY_SIZE, "%s%s@", tag, user);
		TextLower(kbuffer, 0, -1);
		key.data = kbuffer;

		if ((status = smdbAccessGetAccess(sm, &key)) != SMDB_ACCESS_UNKNOWN)
			goto result;
	}
result:	
	free(kbuffer);
error0:	
	return status;
}

/*
 * Lookup
 *
 *	tag:a.b.c.d
 *	tag:a.b.c
 *	tag:a.b
 *	tag:a
 *
 *	a.b.c.d
 *	a.b.c
 *	a.b
 *	a
 *
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessIsIpOk(smdb *sm, const char *tag, char *ip)
{
	int rc;

	if ((rc = smdbAccessGetIpAccess(sm, tag, ip)) != SMDB_ACCESS_UNKNOWN)
		return rc;

	return smdbAccessGetIpAccess(sm, (const char *) 0, ip);
}

/*
 * Lookup
 *
 *	tag:some.sub.domain.tld
 *	tag:sub.domain.tld
 *	tag:domain.tld
 *	tag:tld
 *
 *	some.sub.domain.tld
 *	sub.domain.tld
 *	domain.tld
 *	tld
 *
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessIsDomainOk(smdb *sm, const char *tag, const char *host)
{
	int rc;

	if ((rc = smdbAccessGetDomainAccess(sm, tag, host)) != SMDB_ACCESS_UNKNOWN)
		return rc;

	return smdbAccessGetDomainAccess(sm, (char *) 0, host);
}

/*
 * Lookup
 *
 *	tag:account@some.sub.domain.tld
 *	tag:some.sub.domain.tld
 *	tag:sub.domain.tld
 *	tag:domain.tld
 *	tag:tld
 *	tag:account@
 *
 *	account@some.sub.domain.tld
 *	some.sub.domain.tld
 *	sub.domain.tld
 *	domain.tld
 *	tld
 *	account@
 *
 * Return one of SMDB_ACCESS_UNKNOWN, SMDB_ACCESS_OK, SMDB_ACCESS_REJECT, SMDB_ACCESS_DISCARD.
 */
int
smdbAccessIsEmailDomainUserOk(smdb *sm, const char *tag, const char *email, const char *domain, const char *user)
{
	int rc;

	if ((rc = smdbAccessGetEmailDomainUserAccess(sm, tag, email, domain, user)) != SMDB_ACCESS_UNKNOWN)
		return rc;

	return smdbAccessGetEmailDomainUserAccess(sm, (const char *) 0, email, domain, user);
}

/***********************************************************************
 *** Sendmail cf parse routines.
 ***********************************************************************/

smdb *smdbAccess;
Vector smdbAliases;

static int
parseCfAccessFile(char *line)
{
	Vector words;
	int rc, dbtype;
	char *word, *filename;

	rc = -1;

	if (debug & SMDB_DEBUG_INFO)
		syslog(LOG_DEBUG, "parseCfAccessFile(%s)", line);
	
	words = TextSplit(line, " \t", 0);
	if (words == (Vector) 0) {
		syslog(LOG_ERR, "failed to split line, %s: %s", strerror(errno), line);
		goto error0;
	}

	if ((word = VectorGet(words, VectorLength(words)-1)) == (char *) 0) {
		syslog(LOG_ERR, "failed to get last word");
		goto error1;
	}

	if (TextInsensitiveEndsWith(word, ".db") < 0) {
		int n = strlen(word) + 4;
		if ((word = realloc(word, n)) == (char *) 0) {
			syslog(LOG_ERR, "failed to realloc() word");
			goto error1;
		}
		VectorSet(words, VectorLength(words)-1, word);
		(void) strncat(word, ".db", n);
	}

	if (debug & SMDB_DEBUG_INFO)
		syslog(LOG_DEBUG, "accessFile=%s", word);

	filename = word;
#if DB_VERSION_MAJOR == 1
	if ((word = VectorGet(words, 1)) == (char *) 0) {
		syslog(LOG_ERR, "failed to get database type");
		goto error1;
	}

	if (TextInsensitiveCompare(word, "btree") == 0) {
		dbtype = DB_BTREE;
	} else if (TextInsensitiveCompare(word, "hash") == 0) {
		dbtype = DB_HASH;
	} else {
		syslog(LOG_ERR, "unsupported database type");
		goto error1;
	}
#else
	/* Later versions can determine the database type themselves. */
	dbtype = DB_UNKNOWN;
#endif
	if ((smdbAccess = smdbOpen((const char *) filename, dbtype)) != (smdb *) 0)
		rc = 0;
error1:
	VectorDestroy(words);	
error0:
	return rc;
}

static int
parseCfAliasFile(char *line)
{
	smdb *sm;
	int dbtype;
	char *token;
	const char *next;	

	if (debug & SMDB_DEBUG_INFO)
		syslog(LOG_DEBUG, "parseCfAliasFile(%s)", line);
	
#if DB_VERSION_MAJOR == 1
	dbtype = DB_HASH;
#else
	dbtype = DB_UNKNOWN;
#endif
	next = line + sizeof("O AliasFile=") - 1;

	while ((token = TextToken(next, &next, ", \t", 0)) != (char *) 0) {
		if (TextInsensitiveEndsWith(token, ".db") < 0) {
			int n = strlen(token) + 4;
			if ((token = realloc(token, n)) == (char *) 0) {
				syslog(LOG_ERR, "failed to realloc() token");
				goto error1;
			}
			(void) strncat(token, ".db", n);
		}

		if (debug & SMDB_DEBUG_INFO)
			syslog(LOG_DEBUG, "aliasesFile=%s", token);

		if ((sm = smdbOpen(token, dbtype)) == (smdb *) 0) {
			syslog(LOG_ERR, "failed to open aliasesFile=%s", token);
			goto error1;
		}

		(void) VectorAdd(smdbAliases, sm);
		free(token);
	}
	
	return 0;
error1:
	free(token);
/* error0: */
	return -1;
}

void
smdbRemoveAliasesEntry(void *entry)
{
	smdb *sm = (smdb *) entry;

#if DB_VERSION_MAJOR == 1
	sm->db->close(sm->db);
#else
	sm->db->close(sm->db, 0);
#endif
	free(sm->filename);
	free(sm);
}

#endif /* LIBSNERT_WITH_BERKELEY_DB */

char *smMasqueradeAs;

static int
parseCfGetMacro(char *line, char *macro, char **value)
{
	if (*line == 'D' && 0 < TextSensitiveStartsWith(line+1, macro))
		*value = TextDup(line + 1 + strlen(macro));
	
	return 0;
}

SmClientPortOptions smClientPortInet4;
SmClientPortOptions smClientPortInet6;

static int
parseCfClientPortOptions(char *line)
{
	char *here, *p;
	SmClientPortOptions *cp;
	
	if (debug & SMDB_DEBUG_INFO)
		syslog(LOG_DEBUG, "parseCfClientPortOptions(%s)", line);
		
#ifdef AF_INET6		
	if (strstr(line, "Family=inet6") != (char *) 0) {
		cp = &smClientPortInet6;
		cp->family = AF_INET6;
	} else 
#endif	
	{
		cp = &smClientPortInet4;
		cp->family = AF_INET;
	}

	cp->port = 25;
	cp->useForHelo = 0;
	cp->dontUseAuth = 0;
	cp->dontUseStartTls = 0;
	cp->address = (char *) 0;
	
	if ((here = strstr(line, "Port=")) != (char *) 0)
		/* No support for port by service name. */
		cp->port = strtol(here, (char **) 0, sizeof ("Port"));
		
	if ((here = strstr(line, "Address=")) != (char *) 0)
		cp->address = TextSubstring(here, sizeof ("Address"), strcspn(here + sizeof ("Address"), ","));

	if ((here = strstr(line, "Modifier=")) != (char *) 0) {
		for (p = here + sizeof ("Modifier"); *p != '\0' && *p != ','; p++) {
			switch (*p) {
			case 'h': cp->useForHelo = 1; break;
			case 'A': cp->dontUseAuth = 1; break;
			case 'S': cp->dontUseStartTls = 1; break;
			}
		}
	}		

	return 0;
}

/*
 * Read the sendmail.cf file looking for support files.
 */
int
readSendmailCf(char *cf, long flags)
{
	int rc;
	FILE *fp;
	char *line;

	rc = -1;

	if (cf == (char *) 0 || (fp = fopen(cf, "r")) == (FILE *) 0)
		goto error0;

	if ((line = malloc(SMCF_LINE_SIZE)) == (char *) 0)
		goto error1;

#ifdef LIBSNERT_WITH_BERKELEY_DB
	if ((smdbAliases = VectorCreate(5, smdbRemoveAliasesEntry)) == (Vector) 0)
		goto error2;
#endif /* LIBSNERT_WITH_BERKELEY_DB */		

	while (0 <= TextInputLine(fp, line, SMCF_LINE_SIZE)) {
#ifdef LIBSNERT_WITH_BERKELEY_DB
		if ((flags & SMDB_OPEN_ACCESS) && 0 < TextSensitiveStartsWith(line, "Kaccess")) {
			if (parseCfAccessFile(line))
				goto error2;
			continue;
		}		
		if ((flags & SMDB_OPEN_ALIASES) && 0 < TextSensitiveStartsWith(line, "O AliasFile=")) {
			if (parseCfAliasFile(line))
				goto error2;
			continue;
		}
#endif /* LIBSNERT_WITH_BERKELEY_DB */		
		if (0 < TextSensitiveStartsWith(line, "O ClientPortOptions=")) {
			if (parseCfClientPortOptions(line))
				goto error2;
			continue;
		}

		parseCfGetMacro(line, "M", &smMasqueradeAs);
	}

	if (feof(fp))
		rc = 0;
error2:
	free(line);
error1:
	fclose(fp);
error0:
	return rc;
}


/***********************************************************************
 *** END
 ***********************************************************************/

