/*
 * HashTable.c
 *
 * Copyright 2003, 2004 by Anthony Howe.  All rights reserved.
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <com/snert/lib/crc/Crc.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/util/HashTable.h>

/*
 * The size of the hash table should be a small prime number:
 *
 *	449, 673, 991, 1409, 2411
 *
 * Using a prime number for the table size means that double
 * hashing or linear-probing can visit all possible entries.
 *
 * This is NOT a runtime option, because its not something I
 * want people to play with unless absolutely necessary.
 */
#ifndef TABLE_SIZE
#define TABLE_SIZE		991
#endif

struct entry {
	void *key;
	void *value;
	struct entry *next;
};

struct table {
	long size;
	long count;
	struct entry **base;

	HashFunction onHashFn;

	CompareFunction onCompareFn;

	HashTableWalkFunction onNextFn;
	void **onNextData;

	HashTableRemoveFunction onRemoveFn;
	void **onRemoveData;
};

unsigned long
HashString(void *key)
{
	return hash32((const unsigned char *) key, (int) strlen((const char *) key));
}

int
HashTableWalkIgnore(void *key, void *value, void **data)
{
	/* Do nothing. */
	return WALK_OK;
}

void
HashTableRemoveAbort(void *key, void *value, void **data)
{
	abort();
}

void
HashTableRemoveTrivial(void *key, void *value, void **data)
{
	free(key);
	free(value);
}

void
HashTableOnHash(HashTable A, HashFunction function)
{
	struct table *a = (struct table *) A;

	if (function == (HashFunction) 0) {
		a->onHashFn = HashString;
	} else {
		a->onHashFn = function;
	}
}

void
HashTableOnCompare(HashTable A, CompareFunction function)
{
	struct table *a = (struct table *) A;

	if (function == (CompareFunction) 0) {
		a->onCompareFn = TextInsensitiveCompare;
	} else {
		a->onCompareFn = function;
	}
}

void
HashTableOnNext(HashTable A, HashTableWalkFunction function, void **data)
{
	struct table *a = (struct table *) A;

	if (function == (HashTableWalkFunction) 0) {
		a->onNextFn = HashTableWalkIgnore;
		a->onNextData = (void **) 0;
	} else {
		a->onNextFn = function;
		a->onNextData = data;
	}
}

void
HashTableOnRemove(HashTable A, HashTableRemoveFunction function, void **data)
{
	struct table *a = (struct table *) A;

	if (function == (HashTableRemoveFunction) 0) {
		a->onRemoveFn = HashTableRemoveAbort;
		a->onRemoveData = (void **) 0;
	} else {
		a->onRemoveFn = function;
		a->onRemoveData = data;
	}
}

HashTable
HashTableCreate()
{
	struct table *a;

	a = (struct table *) malloc(sizeof *a);
	if (a == (struct table *) 0)
		return (HashTable) 0;

	a->base = calloc(TABLE_SIZE, sizeof (void *));
	if (a->base == (struct entry **) 0) {
		free(a);
		return (HashTable) 0;
	}

	a->count = 0;
	a->size = TABLE_SIZE;

	HashTableOnHash(a, HashString);
	HashTableOnCompare(a, TextInsensitiveCompare);
	HashTableOnNext(a, HashTableWalkIgnore, (void **) 0);
	HashTableOnRemove(a, HashTableRemoveAbort, (void **) 0);

	return (HashTable) a;

}

static int
HashTableWalkRemove(void *key, void *value, void **data)
{
	return WALK_REMOVE_ENTRY;
}

void
HashTableDestroy(HashTable A)
{
	struct table *a = (struct table *) A;

	if (a != (struct table *) 0) {
		HashTableWalk(a, HashTableWalkRemove, (void **) 0);
		free(a->base);
		free(a);
	}
}

static int
HashTableFetch(HashTable A, void *key, struct entry ***before, struct entry **entry)
{
	struct entry *e, **prev;
	struct table *a = (struct table *) A;
	unsigned long hash = (*a->onHashFn)(key) % a->size;

	for (prev = &a->base[hash], e = *prev; e != (struct entry *) 0; e = *prev) {
		if ((*a->onCompareFn)(key, e->key) == 0) {
			*before = prev;
			*entry = e;
			return 1;
		}

		/* Before skipping an entry, call the onNext function.
		 * This allows us to do expire/garbadge collection like
		 * operations, if we want.
		 */
		switch ((*a->onNextFn)(e->key, e->value, a->onNextData)) {
		case WALK_OK:
			prev = &e->next;
			break;
		case WALK_STOP:
			*before = prev;
			return 0;
		case WALK_REMOVE_ENTRY:
			(*a->onRemoveFn)(e->key, e->value, a->onRemoveData);
			*prev = e->next;
			a->count--;
			free(e);
			break;
		}
	}

	*before = prev;
	return 0;
}

void *
HashTableGet(HashTable A, void *key)
{
	struct entry *e, **prev;

	if (HashTableFetch(A, key, &prev, &e))
		return e->value;

	return (void *) 0;
}

int
HashTableSet(HashTable A, void *key, void *value)
{
	struct entry *e, **prev;

	if (key == (void *) 0)
		return -1;

	if (HashTableFetch(A, key, &prev, &e)) {
		/* Update value of existing key. */
		e->value = value;
		return 0;
	}

	/* Add new key to the end of a hash bucket. */
	e = (struct entry *) malloc(sizeof *e);
	if (e == (struct entry *) 0)
		return -1;

	((struct table *) A)->count++;

	e->key = key;
	e->value = value;
	e->next = *prev;
	*prev = e;

	return 0;
}

void
HashTableRemove(HashTable A, void *key)
{
	struct entry *e, **prev;
	struct table *a = (struct table *) A;

	if (HashTableFetch(A, key, &prev, &e)) {
		(*a->onRemoveFn)(e->key, e->value, a->onRemoveData);
		*prev = e->next;
		a->count--;
		free(e);
	}
}

void
HashTableWalk(HashTable A, HashTableWalkFunction action, void **data)
{
	int i;
	struct entry *e, **prev;
	struct table *a = (struct table *) A;

	for (i = 0; i < a->size; ++i) {
		for (prev = &a->base[i], e = *prev; e != (struct entry *) 0; e = *prev) {
			switch ((*action)(e->key, e->value, data)) {
			case WALK_OK:
				prev = &e->next;
				break;
			case WALK_STOP:
				return;
			case WALK_REMOVE_ENTRY:
				(*a->onRemoveFn)(e->key, e->value, a->onRemoveData);
				*prev = e->next;
				a->count--;
				free(e);
				break;
			}
		}
	}
}

int
HashTableCollectKeys(void *key, void *value, void **data)
{
	Vector vector = (Vector) *data;

	if (VectorAdd(vector, key)) {
		VectorDestroy(vector);
		*data = (void *) 0;
		return WALK_STOP;
	}

	return WALK_OK;
}

int
HashTableCollectValues(void *key, void *value, void **data)
{
	Vector vector = (Vector) *data;

	if (VectorAdd(vector, value)) {
		VectorDestroy(vector);
		*data = (void *) 0;
		return WALK_STOP;
	}

	return WALK_OK;
}

Vector
HashTableKeys(HashTable A)
{
	Vector vector;

	if ((vector = VectorCreate(100, VectorRemoveIgnore)) != (Vector) 0) {
		/* We are coping pointers, not values, so we have
		 * to avoid removing those values until the hash
		 * table is destroyed.
		 */
/*		VectorOnRemove(vector, VectorRemoveIgnore); */
		HashTableWalk(A, HashTableCollectKeys, &vector);
	}

	return vector;
}

Vector
HashTableValues(HashTable A)
{
	Vector vector;

	if ((vector = VectorCreate(100, VectorRemoveIgnore)) != (Vector) 0) {
		/* We are coping pointers, not values, so we have
		 * to avoid removing that data until the hash
		 * table is destroyed.
		 */
/*		VectorOnRemove(vector, VectorRemoveIgnore); */
		HashTableWalk(A, HashTableCollectValues, &vector);
	}

	return vector;
}

long
HashTableCount(HashTable A)
{
	return ((struct table *) A)->count;
}

