/*
 * Pool.c
 *
 * Copyright 2003, 2004 by Anthony Howe.  All rights reserved.
 *
 * This is essentially a lazy man's garbadge collector. Based on
 * chapter 10 from "Obfuscated C and Other Mysteries" by Don Libes,
 * Wiley, 1993.
 */

#include <stdlib.h>

#include <com/snert/lib/util/RefMem.h>

struct pool {
	struct pool *parent;
	unsigned long count;
};

static struct pool *MainPool;

#define POOL_MINIMUM_CHUNK_SIZE		(long)(sizeof (struct pool))

void *
PoolAllocate(struct pool *parent, long size)
{
	struct pool *child;
	
	/* Add space for accounting structure. */
	size += POOL_MINIMUM_CHUNK_SIZE;
	if (size < POOL_MINIMUM_CHUNK_SIZE)
		return (void *) 0;
		
	if ((block = (void *) malloc(size)) == (void *) 0)
		return (void *) 0;

	block->parent = parent;
	block->count = 1;
	
	return (void *)(block + 1);	
}

static void *
_PoolFree(Pool node)
{
	struct pool *here, *next;
	struct pool *block = (struct pool *) node - 1;	
	
	for (here = block->parent; here != (struct pool *) 0; ) {
		
	}
}

void
PoolFree(Pool node)
{
	(void) _PoolFree(node);
}

void
PoolSetParent(Pool parent, Pool node)
{
	struct pool *block = (struct pool *) node - 1;	
	block->parent = (struct pool *) parent;
}


static t_ref *head = (t_ref *) 0;

STATIC t_ref *_reffree _((void *));

/*
 * Allocate a new object if data is null, else resize an existing object.
 * Return pointer to allocated memory, else null if not enough memory.
 */
void *
_refalloc(data, ref, len)
void *data, *ref;
unsigned long len;
{
	void *new;
	t_ref *node, *ptr;

	/* Allocate node first for case when data is not in the list.
	 * If we successfully realloc(data, len) and then failed to
	 * allocate a new node, then there would be no way to "back-out"
	 * of the reallocation of data to return an error condition.
	 * This way be may fragment heap memory, but its more robust.
	 */
	if ((node = (t_ref *) malloc(sizeof *ptr)) == (t_ref *) 0)
		return (t_ptr) 0;

	if ((data == (t_ptr) 0 and (new = malloc(len)) == (t_ptr) 0)
	or (new = realloc(data, len)) == (t_ptr) 0) {
		free(node);
		return (t_ptr) 0;
	}

	if (data != (t_ptr) 0) {
		/* Reallocation, update relavent node in the list. */
		for (ptr = head; ptr != (t_ref *) 0; ptr = ptr->next) {
			if (ptr->data == data) {
				ptr->data = new;
				ptr->ref = ref;
				free(node);
				return new;
			}
		}
	}

	/* Add new or resized data item to list. */
	node->next = head;
	node->data = new;
	node->ref = ref;
	head = node;

	return new;
}

STATIC t_ref *
_reffree(data)
void *data;
{
	t_ref **back, *ptr;

	/* Look for nodes dependent upon data. */
	for (ptr = head; ptr != (t_ref *) 0; ) {
		if (ptr->ref == data) {
			/* Break cycles by removing this reference. */
			ptr->ref = (t_ptr) 0;
			ptr = _reffree(ptr->data);
		} else {
			ptr = ptr->next;
		}
	}

	/* Remove node associated with data from list and free it. */
	for (back = &head, ptr = head; ptr != (t_ref *) 0; ptr = *back) {
		if (ptr->data == data) {
			*back = ptr->next;
			free(data);
			free(ptr);
			break;
		}

		back = &ptr->next;
	}

	/* Return the next node following the deleted one. */
	return *back;
}

/*
 * Always free data and any associated data.
 */
void
reffree(data)
void *data;
{
	(void) _reffree(data);
}

/*
 * Change the reference of a data block.  If the block is not in the
 * reference list, then add it to the list.  Return 0 for success,
 * else -1 if data was not in the list and could not be added because
 * malloc() failed.
 *
 * Note that data block must have been returned from a previous call
 * to refalloc() or malloc(), because adding a static data block to the
 * reference list will have undefined results for reffree().
 */
int
refchange(data, ref)
void *data, *ref;
{
	t_ref *ptr;

	for (ptr = head; ptr != (t_ref *) 0; ptr = ptr->next) {
		if (ptr->data == data) {
			ptr->ref = ref;
			return 0;
		}
	}

	/* Data is not in reference list, add it. */
	if ((ptr = (t_ref *) malloc(sizeof *ptr)) == (t_ref *) 0)
		return -1;

	ptr->next = head;
	ptr->data = data;
	ptr->ref = ref;
	head = ptr;

	return 0;
}

/*
 * Print list of unfreed memory blocks.
 * This function suitable for atexit().
 */
void
refdebug()
{
	t_ref *ptr;

	for (ptr = head; ptr != (t_ref *) 0; ptr = ptr->next)
		error(__FILE__, __LINE__, "Garbage %p.", ptr->data);
}
