/*
 * Vector.c
 *
 * Dynamic Vector API
 *
 * Copyright 2001, 2004 by Anthony Howe.  All rights reserved.
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <com/snert/lib/util/Vector.h>

#ifndef VECTOR_GROWTH
#define VECTOR_GROWTH		1000
#endif

struct vector {
	void (*destroy)(Vector);
	void (*onRemoveFn)(void *);
	long capacity;		/* Number of allocated elements. */
	long length;		/* Number of used elements. */
	void **base;		/* Base of the array of elements. */
};

Vector
VectorCreate(long capacity, void (*destroy)(void *entry))
{
	struct vector *a;
	
	if (capacity < 1)
		capacity = 1;

	a = (struct vector *) malloc(sizeof *a);
	if (a == (struct vector *) 0)
		return (Vector) 0;

	/* Allocate an extra element so that the vector can
	 * maintain a null object at the end. The vector is
	 * an array of pointers, so a null pointer at the end
	 * can be used for an end condition when the vector
	 * is iterated over.
	 */	 
	a->base = malloc(sizeof (void *) * (capacity + 1));
	
	if (a->base == (void **) 0) {
		free(a);
		return (Vector) 0;
	}

	VectorOnRemove(a, destroy);
	*a->base = (void *) 0;
	a->capacity = capacity;
	a->length = 0;

	return (Vector) a;
}

void
VectorDestroy(Vector A)
{
	struct vector *a = (struct vector *) A;

	if (a != (struct vector *) 0) {
		VectorRemove(a, 0, a->length);
		free(a->base);
		free(a);
	}
}

long
VectorCapacity(Vector A)
{
	return ((struct vector *) A)->capacity;
}

long
VectorLength(Vector A)
{
	return ((struct vector *) A)->length;
}

void **
VectorBase(Vector A)
{
	return ((struct vector *) A)->base;
}

int
VectorAdd(Vector A, void *data)
{
	void **bigger;
	struct vector *a = (struct vector *) A;

	if (a->capacity <= a->length) {
		bigger = realloc(a->base, sizeof (void *) * (a->capacity + VECTOR_GROWTH + 1));
		if (bigger == (void **) 0)
			return -1;
		a->capacity += VECTOR_GROWTH;
		a->base = bigger;
	}

	a->base[a->length++] = data;
	a->base[a->length] = (void *) 0;

	return 0;
}

void *
VectorGet(Vector A, long index)
{
	if (0 <= index && index < ((struct vector *) A)->length)
		return ((struct vector *) A)->base[index];

	return (void *) 0;
}

void
VectorSet(Vector A, long index, void *data)
{
	if (0 <= index && index < ((struct vector *) A)->length)
		((struct vector *) A)->base[index] = data;
}

void
VectorSort(Vector A, CompareFunction cmp)
{
	if (A == (Vector) 0)
		return;

	qsort(
		((struct vector *) A)->base,
		((struct vector *) A)->length,
		sizeof (void *), cmp
	);
}

/*
 * Reverse range x..y inclusive.
 */
void
VectorReverse(Vector A, long x, long y)
{
	void *temp;
	void **xp = &((struct vector *) A)->base[x];
	void **yp = &((struct vector *) A)->base[y];

	while (xp < yp) {
		temp = *xp;
		*xp++ = *yp;
		*yp-- = temp;
	}
}

/*
 * Move range start..finish inclusive before the given location.
 * Return -1 if the region to move is out of bounds or the target
 * falls within the region; 0 for success.
 *
 * (See Software Tools chapter 6.)
 */
int
VectorMove(Vector A, long start, long finish, long to)
{
	struct vector *a = (struct vector *) A;

	if (finish < start || a->length <= finish)
		return -1;

	if (to < start) {
		VectorReverse(A, to, start-1);
		VectorReverse(A, start, finish);
		VectorReverse(A, to, finish);
	} else if (finish < to && to <= a->length) {
		VectorReverse(A, start, finish);
		VectorReverse(A, finish+1, to-1);
		VectorReverse(A, start, to-1);
	} else {
		return -1;
	}

	return 0;
}

void
VectorRemove(Vector A, long index, long length)
{
	long i, j;
	struct vector *a = (struct vector *) A;

	if (a->length < index + length)
		length = a->length - index;

	/* Apply the object removal callback. */
	for (i = index, j = index + length; i < j; i++)
		(*a->onRemoveFn)(a->base[i]);

	/* Move vacant slots to end of array. */
	if (VectorMove(A, index, index + length - 1, a->length) == 0)
		a->length -= length;
}

void
VectorWalk(Vector A, VectorWalkFunction action, void *data)
{
	long i;
	struct vector *a = (struct vector *) A;

	for (i = a->length; 0 <= --i; ) {
		switch ((*action)(a->base[i], data)) {
		case WALK_OK:
			break;
		case WALK_STOP:
			return;
		case WALK_REMOVE_ENTRY:
			VectorRemove(a, i, 1);
			a->length--;
			break;
		}
	}
}

int
VectorWalkIgnore(void *entry, void *data)
{
	/* Do nothing. */

	return WALK_OK;
}

void
VectorRemoveAbort(void *entry)
{
	abort();
}

void
VectorRemoveIgnore(void *entry)
{
	/* Do nothing. */
}

void
VectorOnRemove(Vector A, void (*destroy)(void *entry))
{
	if (destroy == (void (*)(void *)) 0)
		destroy = VectorRemoveIgnore;		

	((struct vector *) A)->onRemoveFn = destroy;
}
