/*
 * show.c
 *
 * Copyright 2000, 2003 by Anthony Howe.  All rights reserved.
 *
 * usage: show lines files...
 */

#include <com/snert/lib/version.h>

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <com/snert/lib/io/posix.h>
#include <com/snert/lib/util/getopt.h>

char buf[BUFSIZ];

char usage[] =
"\033[1musage: show [-f][-p pat] count files...\033[0m\n"
"\n"
"-f\t\tcontinue to output data as the file grows\n"
"-p pat\t\thighlight pattern in the output\n"
"count\t\tDisplay the top (+ve) or bottom (-ve) N lines; 0 for all\n"
"files ...\tlist of files to show\n"
"\n"
"\033[1mshow/1.2 Copyright 2000, 2004 by Anthony Howe. All rights reserved.\033[0m\n"
;

const char ansiNormal[] = "\033[0m";
const char ansiReverse[] = "\033[7m";

int
output(const char *buf, long length, const char *pattern, long *resume)
{
	long i, plen;
	const char *stop;
	
	if (pattern == (const char *) 0)
		return write(1, buf, length);

	plen = strlen(pattern);	
	stop = buf + length;
	
	if (0 < *resume) {	
		for (i = 0; buf + i < stop; i++) {
			if (buf[i] != pattern[*resume + i])
				break;
		}
		
		if (pattern[*resume + i] == '\0') {
			write(1, ansiReverse, sizeof (ansiReverse) - 1);
			write(1, pattern, plen);
			write(1, ansiNormal, sizeof (ansiNormal) - 1);
			buf += i;
		} else if (stop <= buf + i) {
			*resume += i;
			return length - i;
		} else {
			write(1, pattern, *resume + i);
			buf += i;
		}

		*resume = 0;
	}

	while (buf < stop) {
		for (i = 0; buf + i < stop; i++) {
			if (buf[i] == pattern[0])
				break;
		}
			
		(void) write(1, buf, i);
		buf += i;
				
		for (i = 0; buf + i < stop && pattern[i] != '\0'; i++) {
			if (buf[i] != pattern[i])
				break;
		}
		
		if (pattern[i] == '\0') {
			write(1, ansiReverse, sizeof (ansiReverse) - 1);
			write(1, pattern, plen);
			write(1, ansiNormal, sizeof (ansiNormal) - 1);
			buf += i;
		} else if (stop <= buf + i) {
			*resume += i;
			return length - i;
		} else {
			write(1, buf, i);
			buf += i;
		}		
	}
	
	return length;
}

int
main(int argc, char **argv)
{
	size_t n;
	off_t offset;
	struct stat finfo;
	int ch, fd, follow;
	char *b, *eb, *pattern;
	long lines, count, patternOffset;
	
	follow = 0;
	pattern = NULL;
	patternOffset = 0;	

	while ((ch = getopt(argc, argv, "fp:1234567890")) != -1) {
		if (isdigit(ch)) {
			optind--;
			break;
		}
			
		switch (ch) {
		case 'f':
			follow = 1;
			break;
		case 'p':
			pattern = optarg;
			break;
		default:
			fprintf(stderr, usage);
			return 2;
		}
	}

	if (optind + 2 != argc) {
		fprintf(stderr, usage);
		return 2;
	}

	lines = strtol(argv[optind++], (char **) 0, 10);

	for ( ; optind < argc; optind++) {
		if (stat(argv[optind], &finfo) < 0)
			return 1;

		if ((fd = open(argv[optind], O_RDONLY)) < 0)
			return 1;

		if (lines == 0) {
			while (0 < (n = read(fd, buf, sizeof (buf)))) {
				(void) output(buf, n, pattern, &patternOffset);
			}
		} else {
			if (lines < 0) {
				count = lines = -lines;

				/* Start with the odd buffer length from the end of file. */
				offset = finfo.st_size % sizeof (buf);

				if ((offset = lseek(fd, -offset, SEEK_END)) == (off_t) -1)
					return 1;

				for (;;) {
					/* Fill the buffer. */
					if ((n = read(fd, buf, sizeof (buf))) == (size_t) -1)
						break;

					/* Count backwards newlines (fine for Unix, PC). */
					for (eb = buf+n; buf < eb; ) {
						if (*--eb == '\n' && --count < 0) {
							offset += (eb+1-buf);
							(void) lseek(fd, offset, SEEK_SET);
							break;
						}
					}

					if (count <= 0)
						break;

					/* Move backwards by BUFSIZ byte units until we
					 * seek beyond the start of the file.
					 */
					if ((offset = lseek(fd, offset-sizeof (buf), SEEK_SET)) == (off_t) -1)
						break;
				}
			}

			while (0 < lines) {
				if ((n = read(fd, buf, sizeof (buf))) < 0)
					break;

				/* Count newlines (fine for Unix, PC). */
				for (b = buf, eb = buf+n; b < eb; )
					if (*b++ == '\n' && --lines <= 0)
						break;

				(void) output(buf, b - buf, pattern, &patternOffset);
			}
		}
		
		while (follow && 0 <= (n = read(fd, buf, sizeof (buf)))) {
			(void) output(buf, n, pattern, &patternOffset);
		}

		(void) close(fd);
	}

	return 0;
}
